//
//  UICollectionView.swift
//  SwiftPokedex
//
//  Created by Viktor Gidlöf on 2021-05-04.
//

import UIKit

extension UICollectionView {
    static let footer = UICollectionView.elementKindSectionFooter
    
    struct Section {
        var title: String? = nil
        let items: [CollectionCellConfigurator]
    }

    struct DataSource {
        var sections = [Section]()
    }

    func cell<Cell: UICollectionViewCell>(at indexPath: IndexPath) -> Cell? {
        cellForItem(at: indexPath) as? Cell
    }
    
    func registerCell<Cell: UICollectionViewCell>(_ cell: Cell.Type) {
        register(Cell.self, forCellWithReuseIdentifier: reuseIdentifier(for: cell))
    }

    func registerReusableFooter<View: UICollectionReusableView>(view: View.Type) {
        register(View.self, forSupplementaryViewOfKind: UICollectionView.footer, withReuseIdentifier: String(describing: View.self))
    }
    
    func dequeueReusableView<View: UICollectionReusableView>(ofKind kind: String, at indexPath: IndexPath) -> View {
        dequeueReusableSupplementaryView(ofKind: kind, withReuseIdentifier: String(describing: View.self), for: indexPath) as! View
    }
    
    func dequeueCell<Cell: UICollectionViewCell>(for item: CollectionCellConfigurator, at indexPath: IndexPath) -> Cell {
        dequeueReusableCell(withReuseIdentifier: type(of: item).reuseId, for: indexPath) as! Cell
    }

    private func reuseIdentifier<Cell: UICollectionViewCell>(for cell: Cell.Type) -> String {
        String(describing: cell)
    }
}

extension UICollectionView.DataSource {
    
    var hasData: Bool { !sections.isEmpty }
    
    func numberOfSections() -> Int {
        sections.count
    }
    
    func numberOfItems(in section: Int) -> Int {
        guard section < sections.count else { return 0 }
        return sections[section].items.count
    }
    
    func item(at indexPath: IndexPath) -> CollectionCellConfigurator {
        sections[indexPath.section].items[indexPath.row]
    }
}

extension UICollectionView.DataSource {
    
    static func pokemonDataSource(from pokemon: [PokemonDetails]) -> UICollectionView.DataSource {
        let sorted = pokemon.sorted(by: { $0.id < $1.id })
        let items: [CollectionCell] = sorted.map { .pokemonCell(from: $0) }
        let section = UICollectionView.Section(items: items)
        let collectionData = UICollectionView.DataSource(sections: [section])
        return collectionData
    }
}
